#include <gmp.h>
#include <mpfr.h>
#include <fplll/fplll.h>
#include "lll_uses.hh"

const double kernel_scaling = 1e18; // TODO coeffs

FF matrix_norm_2(const ff_matrix& m) {
  FF norm = 0.0;
  for(unsigned i=0; i<m.size1(); i++) {
    matrix_row<const ff_matrix> r(m, i);
    for(matrix_row<const ff_matrix>::const_iterator pos = r.begin();
	pos != r.end(); pos++) {
      FF c = *pos;
      norm += c*c;
    }
  }
  return norm;
}

qq_vector left_near_kernel_vector(const ff_matrix& m) {
  const double scaling = kernel_scaling/matrix_norm_2(m);
  ZZ_mat<mpz_t> mlat(m.size1(), m.size2()+m.size1());
  FP_NR<double> f;
  Z_NR<mpz_t> z;

  for(unsigned i=0; i<m.size1(); i++) {
    matrix_row<const ff_matrix> rowi(m, i);
    for(matrix_row<const ff_matrix>::const_iterator pos=rowi.begin();
	pos != rowi.end(); pos++) {
      //std::cout << "mat[" << i << ", " << pos.index() << "] = " << *pos << std::endl;
      f.set(*pos * scaling);
      z.set_f(f);
      mlat.Set(i, pos.index(), z); 
    }
  }
  z.set(1);
  for(unsigned i=0; i<m.size1(); i++) {
    mlat.Set(i, m.size2()+i, z);
  }
  wrapper(&mlat).LLL();
  qq_vector ret(m.size1());
#if 1
  for(unsigned i=0; i<m.size1(); i++) {
    FastInteger x(mlat.Get(0, m.size2()+i).GetData());
    if (x.isZero()) continue;
     ret[i] = FastRational(x);
  }
  std::cout << "kernel vector norm: " << norm_2(prod(qq_to_ff_vector(ret), m)) << std::endl;
#else
  for(int i=m.size1()-1; i>=0; i--) {
    for(unsigned j=0; j<m.size1(); j++) {
      FastInteger x(mlat.Get(i, m.size2()+j).GetData());
      ret[j] = FastRational(x);
    }
    ff_vector ret_f = qq_to_ff_vector(ret);
    std::cout << "kernel vector norm: " << norm_1(ret_f) << " image norm: " << norm_2(prod(ret_f, m)) << std::endl;
  }
#endif
  return ret;
}

void left_near_kernel_vectors(const ff_matrix& m, std::vector<qq_vector>& dest) {
  const double scaling = kernel_scaling/matrix_norm_2(m);
  std::cout << "LLL scaling: " << scaling << std::endl;

  ZZ_mat<mpz_t> mlat(m.size1(), m.size2()+m.size1());
  FP_NR<double> f;
  Z_NR<mpz_t> z;

  for(unsigned i=0; i<m.size1(); i++) {
    matrix_row<const ff_matrix> rowi(m, i);
    for(matrix_row<const ff_matrix>::const_iterator pos=rowi.begin();
	pos != rowi.end(); pos++) {
      //std::cout << "mat[" << i << ", " << pos.index() << "] = " << *pos << std::endl;
      f.set(*pos * scaling);
      z.set_f(f);
      mlat.Set(i, pos.index(), z); 
    }
  }
  z.set(1);
  for(unsigned i=0; i<m.size1(); i++) {
    mlat.Set(i, m.size2()+i, z);
  }
  wrapper(&mlat).LLL();
  qq_vector ret(m.size1());

  for(unsigned j=0; j<m.size1(); j++) {
    FastInteger x(mlat.Get(0, m.size2()+j).GetData());
    ret[j] = FastRational(x);
  }
  dest.push_back(ret);
  ff_vector ret_f = qq_to_ff_vector(ret);
  // TODO coeffs
  double norm_v = norm_1(ret_f), norm_image = norm_2(prod(ret_f, m));
  std::cout << "leading kernel vector: norm: " << norm_v << ", image norm: " << norm_image << std::endl;
  double thresh_norm_v = norm_v * 10., thresh_norm_image = norm_image * 10.;

  for(unsigned i=1; i<m.size1(); i++) {
    for(unsigned j=0; j<m.size1(); j++) {
      FastInteger x(mlat.Get(i, m.size2()+j).GetData());
      ret[j] = FastRational(x);
    }
    ret_f = qq_to_ff_vector(ret);
    norm_v = norm_1(ret_f);
    if (norm_v > thresh_norm_v) continue;
    norm_image = norm_2(prod(ret_f, m));
    if (norm_image > thresh_norm_image) continue;

    std::cout << "kernel vector #" << i << ": norm: " << norm_v << ", image norm: " << norm_image << std::endl;
    dest.push_back(ret);
  }
}

#if 1
qq_vector near_vector(double magnitude, const ff_vector& v,
		      const ff_vector& f0, const ff_matrix& fis) {
  assert(f0.size() == fis.size2());
  ff_vector delta = f0-v;
  unsigned dim = f0.size(), search_dim = fis.size1();
  double beta = 1e18; // TODO
  double alpha = beta * magnitude;
  ZZ_mat<mpz_t> mlat(search_dim+1, 2*dim+search_dim+1);
  FP_NR<double> f;
  Z_NR<mpz_t> z;
  for(ff_vector::const_iterator pos=delta.begin(); pos!=delta.end(); pos++) {
    f.set(alpha * (*pos));
    z.set_f(f);
    mlat.Set(0, pos.index(), z);
  }
  for(ff_vector::const_iterator pos=f0.begin(); pos!=f0.end(); pos++) {
    f.set(beta * (*pos));
    z.set_f(f);
    mlat.Set(0, dim+pos.index(), z);
  }
  for(unsigned i=0; i<search_dim; i++) {
    matrix_row<const ff_matrix> rowi(fis, i);
    for(matrix_row<const ff_matrix>::const_iterator pos=rowi.begin();
	pos != rowi.end(); pos++) {
      f.set(alpha * (*pos));
      z.set_f(f);
      mlat.Set(i+1, pos.index(), z);

      f.set(beta * (*pos));
      z.set_f(f);
      mlat.Set(i+1, dim+pos.index(), z);
    }
  }
  z.set(1);
  for(unsigned i=0; i<=search_dim; i++) {
    mlat.Set(i, 2*dim+i, z);
  }
  wrapper(&mlat).LLL();
  for(unsigned i=0; i<search_dim; i++) {
    ZZ denominator(mlat.Get(i, 2*dim).GetData());
    if(denominator.isZero()) continue;
    std::cout << "denominator = " << denominator << std::endl;

    qq_vector ret(search_dim);
    for(unsigned j=0; j<search_dim; j++) {
      ZZ numerator(mlat.Get(i, 2*dim+1+j).GetData());
      if (numerator.isZero()) continue;
      ret[j] = QQ(numerator) / denominator;
    }
    return ret;
  }
  return qq_vector(0);
}
#else

qq_vector near_vector(double magnitude, const ff_vector& v,
		      const ff_vector& f0, const ff_matrix& fis) {
  assert(f0.size() == fis.size2());
  ff_vector delta = f0-v;
  unsigned dim = f0.size(), search_dim = fis.size1();
  double beta = 1e18; // TODO
  double alpha = beta * magnitude;
  ZZ_mat<mpz_t> mlat(search_dim+1, dim+search_dim+2);
  FP_NR<double> f;
  Z_NR<mpz_t> z;
  for(ff_vector::const_iterator pos=delta.begin(); pos!=delta.end(); pos++) {
    f.set(alpha * (*pos));
    z.set_f(f);
    mlat.Set(0, pos.index(), z);
  }
  for(unsigned i=0; i<search_dim; i++) {
    matrix_row<const ff_matrix> rowi(fis, i);
    for(matrix_row<const ff_matrix>::const_iterator pos=rowi.begin();
	pos != rowi.end(); pos++) {
      f.set(alpha * (*pos));
      z.set_f(f);
      mlat.Set(i+1, pos.index(), z);
    }
  }
  f.set(beta);
  z.set_f(f);
  mlat.Set(0, dim, z);
  z.set(1);
  for(unsigned i=0; i<=search_dim; i++) {
    mlat.Set(i, dim+1+i, z);
  }
  wrapper(&mlat).LLL();
  for(unsigned i=0; i<search_dim; i++) {
    ZZ denominator(mlat.Get(i, dim+1).GetData());
    if(denominator.isZero()) continue;
    std::cout << "denominator = " << denominator << std::endl;

    qq_vector ret(search_dim);
    for(unsigned j=0; j<search_dim; j++) {
      ZZ numerator(mlat.Get(i, dim+2).GetData());
      if (numerator.isZero()) continue;
      ret[j] = QQ(numerator) / denominator;
    }
    return ret;
  }
  return qq_vector(0);
}
#endif
